==> js-20.0.0-sources.jar: com/oracle/truffle/js/nodes/access/GetPrototypeFromConstructorNode.java <==
/*
 * Copyright (c) 2018, 2018, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.oracle.truffle.js.nodes.access;

import com.oracle.truffle.api.frame.VirtualFrame;
import com.oracle.truffle.api.object.DynamicObject;
import com.oracle.truffle.js.nodes.JavaScriptNode;
import com.oracle.truffle.js.runtime.JSContext;
import com.oracle.truffle.js.runtime.JSRealm;
import com.oracle.truffle.js.runtime.JSRuntime;
import com.oracle.truffle.js.runtime.objects.JSObject;
import com.oracle.truffle.js.runtime.util.CompilableFunction;

public class GetPrototypeFromConstructorNode extends JavaScriptNode {
    private final CompilableFunction<JSRealm, DynamicObject> intrinsicDefaultProto;
    @Child private JavaScriptNode constructorNode;
    @Child private PropertyGetNode getPrototypeNode;
    @Child private IsJSObjectNode isObjectNode;

    protected GetPrototypeFromConstructorNode(JSContext context, JavaScriptNode constructorNode, CompilableFunction<JSRealm, DynamicObject> intrinsicDefaultProto) {
        this.constructorNode = constructorNode;
        this.intrinsicDefaultProto = intrinsicDefaultProto;
        this.getPrototypeNode = PropertyGetNode.create(JSObject.PROTOTYPE, false, context);
        this.isObjectNode = IsJSObjectNode.create();
    }

    public static GetPrototypeFromConstructorNode create(JSContext context, JavaScriptNode constructorNode, CompilableFunction<JSRealm, DynamicObject> intrinsicDefaultProto) {
        return new GetPrototypeFromConstructorNode(context, constructorNode, intrinsicDefaultProto);
    }

    @Override
    public final DynamicObject execute(VirtualFrame frame) {
        return executeDynamicObject(frame);
    }

    @Override
    public DynamicObject executeDynamicObject(VirtualFrame frame) {
        Object constructor = constructorNode.execute(frame);
        return executeWithConstructor((DynamicObject) constructor);
    }

    public DynamicObject executeWithConstructor(DynamicObject constructor) {
        assert JSRuntime.isCallable(constructor);
        Object proto = getPrototypeNode.getValue(constructor);
        if (isObjectNode.executeBoolean(proto)) {
            assert JSRuntime.isObject(proto);
            return (DynamicObject) proto;
        } else {
            JSRealm realm = JSRuntime.getFunctionRealm(constructor, getPrototypeNode.getContext());
            return intrinsicDefaultProto.apply(realm);
        }
    }

    @Override
    public boolean isResultAlwaysOfType(Class<?> clazz) {
        return clazz == DynamicObject.class;
    }

    @Override
    protected JavaScriptNode copyUninitialized() {
        return new GetPrototypeFromConstructorNode(getPrototypeNode.getContext(), cloneUninitialized(constructorNode), intrinsicDefaultProto);
    }
}

==> js-22.0.0.2-sources.jar: com/oracle/truffle/js/nodes/access/GetPrototypeFromConstructorNode.java <==
/*
 * Copyright (c) 2018, 2021, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.oracle.truffle.js.nodes.access;

import com.oracle.truffle.api.frame.VirtualFrame;
import com.oracle.truffle.api.instrumentation.Tag;
import com.oracle.truffle.api.object.DynamicObject;
import com.oracle.truffle.js.nodes.JavaScriptNode;
import com.oracle.truffle.js.runtime.JSContext;
import com.oracle.truffle.js.runtime.JSRealm;
import com.oracle.truffle.js.runtime.JSRuntime;
import com.oracle.truffle.js.runtime.objects.JSObject;
import com.oracle.truffle.js.runtime.util.CompilableFunction;

import java.util.Set;

public class GetPrototypeFromConstructorNode extends JavaScriptNode {
    private final CompilableFunction<JSRealm, DynamicObject> intrinsicDefaultProto;
    @Child private JavaScriptNode constructorNode;
    @Child private PropertyGetNode getPrototypeNode;
    @Child private IsJSObjectNode isObjectNode;

    protected GetPrototypeFromConstructorNode(JSContext context, JavaScriptNode constructorNode, CompilableFunction<JSRealm, DynamicObject> intrinsicDefaultProto) {
        this.constructorNode = constructorNode;
        this.intrinsicDefaultProto = intrinsicDefaultProto;
        this.getPrototypeNode = PropertyGetNode.create(JSObject.PROTOTYPE, false, context);
        this.isObjectNode = IsJSObjectNode.create();
    }

    public static GetPrototypeFromConstructorNode create(JSContext context, JavaScriptNode constructorNode, CompilableFunction<JSRealm, DynamicObject> intrinsicDefaultProto) {
        return new GetPrototypeFromConstructorNode(context, constructorNode, intrinsicDefaultProto);
    }

    @Override
    public DynamicObject execute(VirtualFrame frame) {
        Object constructor = constructorNode.execute(frame);
        return executeWithConstructor((DynamicObject) constructor);
    }

    public DynamicObject executeWithConstructor(DynamicObject constructor) {
        assert JSRuntime.isCallable(constructor);
        Object proto = getPrototypeNode.getValue(constructor);
        if (isObjectNode.executeBoolean(proto)) {
            assert JSRuntime.isObject(proto);
            return (DynamicObject) proto;
        } else {
            JSRealm realm = JSRuntime.getFunctionRealm(constructor, getRealm());
            return intrinsicDefaultProto.apply(realm);
        }
    }

    @Override
    public boolean isResultAlwaysOfType(Class<?> clazz) {
        return clazz == DynamicObject.class;
    }

    @Override
    protected JavaScriptNode copyUninitialized(Set<Class<? extends Tag>> materializedTags) {
        return new GetPrototypeFromConstructorNode(getPrototypeNode.getContext(), cloneUninitialized(constructorNode, materializedTags), intrinsicDefaultProto);
    }
}
